#ifndef __INSTRUCTION_H__
#define __INSTRUCTION_H__

#include "AsmBuilder.h"
#include "Operand.h"
#include <map>
#include <sstream>
#include <vector>

static std::string instName[] = {"BINARY",  "COND", "UNCOND", "RET",  "LOAD",
                                 "STORE",   "CMP",  "ALLOCA", "CALL", "ZEXT",
                                 "XOR",     "GEP",  "PHI",
                                 "FPTOSI", // floating point to signed int
                                 "SITOFP", // signed int to floating point
                                 "BITCAST", "SHL",  "ASHR"};

class BasicBlock;
class Operand;

class Instruction {
  public:
    bool flag = 0;
    void setFlag(bool f) { flag = f; }
    bool getFlag() { return flag; }
    enum INSTTYPE {
        BINARY,
        COND,
        UNCOND,
        RET,
        LOAD,
        STORE,
        CMP,
        ALLOCA,
        CALL,
        ZEXT,
        XOR,
        GEP,
        PHI,
        FPTOSI, // floating point to signed int
        SITOFP, // signed int to floating point
        BITCAST,
        SHL,
        ASHR
    };

    std::string getInstName() { return instName[instType]; }
    void push_back_opse(SymbolEntry *se) { op_se.emplace_back(se); }
    virtual void replaceUse(Operand *old, Operand *new_) {}
    virtual void replaceDef(Operand *new_) {}
    Instruction(unsigned instType, BasicBlock *insert_bb = nullptr);
    virtual ~Instruction();
    BasicBlock *getParent();
    unsigned getOpcode() const { return opcode; };
    unsigned getInstType() const { return instType; };
    bool isUncond() const { return instType == UNCOND; };
    bool isCond() const { return instType == COND; };
    bool isAlloc() const { return instType == ALLOCA; };
    bool isRet() const { return instType == RET; };
    bool isCall() const { return instType == CALL; }
    bool isStore() const { return instType == STORE; }
    bool isZext() const { return instType == ZEXT; }
    bool isPhi() const { return instType == PHI; };
    bool isBin() const { return instType == BINARY; };
    bool isLoad() const { return instType == LOAD; };
    bool isCmp() const { return instType == CMP; };
    bool isGep() const { return instType == GEP; };
    bool isXor() const { return instType == XOR; };
    bool isSitofp() const { return instType == SITOFP; };
    bool isBitcast() const { return instType == BITCAST; };
    bool isShl() const { return instType == SHL; };
    bool isVoid() {
        return (instType == RET) || (instType == COND) ||
               (instType == UNCOND) || (instType == STORE) ||
               (instType == CALL);
    };
    bool isStof() { return instType == SITOFP; };
    bool isFtos() { return instType == FPTOSI; };
    void setParent(BasicBlock *);
    void setNext(Instruction *);
    void setPrev(Instruction *);
    Instruction *getNext();
    Instruction *getPrev();
    std::vector<Operand *> &getOp() { return operands; };
    std::vector<SymbolEntry *> &getOpse() { return op_se; };
    void replace_opse(SymbolEntry *new_op,SymbolEntry *old_se){
      for(int i =0;i<op_se.size();i++){
        if(op_se[i]==old_se){
          op_se[i]=new_op;
        }
      }
    }


    virtual void output() const = 0;
    virtual Instruction *copy(Instruction *ins, BasicBlock *new_bb) = 0;
    // {
    //     return new Instruction(insttype, new_bb);
    // }
    MachineOperand *genMachineOperand(Operand *);
    MachineOperand *genMachineFloatOperand(Operand *);
    MachineOperand *genMachineReg(int reg);
    MachineOperand *genMachineFReg(int freg);
    MachineOperand *genMachineVReg(bool fpu = false);
    MachineOperand *genMachineImm(int val);
    MachineOperand *genMachineLabel(int block_no);
    virtual void genMachineCode(AsmBuilder *) = 0;
    // Z0323
    Operand *getDef() {
        if (!operands.empty())
            return operands[0];
        return nullptr;
    };
    // void replaceDef(Operand* new_op) {
    //     if(!operands.empty()){
    //         operands[0]->removeDef(this);
    //         operands[0] = new_op;
    //         new_op->setDef(this);
    //     }
    // }

    virtual Operand *getDst() { return nullptr; };
    virtual std::vector<Operand *> getUse() { return std::vector<Operand *>(); }
    virtual Instruction *copy1() = 0;
    void setOperand(int index, Operand *ope) {
        // std::cout<<"index = "<<index<<" size = "<<operands.size()<<"\n";
        assert(index >= 0 && index < operands.size() &&
               "index too small or too big");
        operands[index] = ope;
    }

  protected:
    unsigned instType;
    unsigned opcode;
    Instruction *prev;
    Instruction *next;
    BasicBlock *parent;
    std::vector<Operand *> operands;
    std::vector<SymbolEntry *> op_se;
    // enum {
    //     BINARY,
    //     COND,
    //     UNCOND,
    //     RET,
    //     LOAD,
    //     STORE,
    //     CMP,
    //     ALLOCA,
    //     CALL,
    //     ZEXT,
    //     XOR,
    //     GEP,
    //     PHI,
    //     FPTOSI,  // floating point to signed int
    //     SITOFP,  // signed int to floating point
    //     BITCAST,
    //     SHL,
    //     ASHR,
    // };
};

// meaningless instruction, used as the head node of the instruction list.
class DummyInstruction : public Instruction {
  public:
    DummyInstruction() : Instruction(-1, nullptr){};
    void output() const {};
    void genMachineCode(AsmBuilder *){};
    void replaceUse(Operand *old, Operand *new_){};
    Instruction *copy1() { return nullptr; }
    Operand *getDst() { return nullptr; };
    Instruction *copy(Instruction *ins, BasicBlock *new_bb) {
        return new DummyInstruction();
    }
};

class AllocaInstruction : public Instruction {
  public:
    AllocaInstruction(Operand *dst, SymbolEntry *se,
                      BasicBlock *insert_bb = nullptr);
    ~AllocaInstruction();
    void output() const;
    void genMachineCode(AsmBuilder *);
    void replaceDef(Operand *new_);
    Operand *getDef() { return operands[0]; }
    // bool isArray() { return se->getType()->isArray();}
    SymbolEntry *getSymbolEntry() { return se; };
    // void replaceSymbolEntry(SymbolEntry* _se) { //函数内联；
    //     this->se = _se;
    // };
    void replaceUse(Operand *old, Operand *new_){};
    Instruction *copy1() { return new AllocaInstruction(*this); }
    Operand *getDst() { return operands[0]; };
    Instruction *copy(Instruction *ins, BasicBlock *new_bb);
    // {
    //     AllocaInstruction* allocains = dynamic_cast<AllocaInstruction*>(ins);
    //     SymbolEntry* op_se = allocains->getOp()[0]->getSymbolEntry();
    //     SymbolEntry* se = allocains->getSymbolEntry();
    //     return new AllocaInstruction(new Operand(op_se), se, new_bb);
    // }

  private:
    SymbolEntry *se;
};

class LoadInstruction : public Instruction {
    private:
        bool loadchange=0;
        int asmtype = 0;// 初值为0，正常生成汇编。如果赋值为1，生成右端为常数的向量化后的汇编。如果赋值为2，生成右端为数组的向量化后的汇编。
   public:
    LoadInstruction(Operand* dst,
                    Operand* src_addr,
                    BasicBlock* insert_bb = nullptr);
    ~LoadInstruction();
    bool loadChange() { return loadchange; }
    void setLoadChange() { loadchange = 1; }
    void output() const;
    void genMachineCode(AsmBuilder *);
    Operand *getDef() { return operands[0]; }
    void replaceUse(Operand *old, Operand *new_);
    void replaceDef(Operand *new_);
    Operand *getAddr() { return this->operands[1]; };
    std::vector<Operand *> getUse() {
        return std::vector<Operand *>({operands[1]});
    }
    Instruction *copy1() { return new LoadInstruction(*this); }
    Operand *getDst() { return operands[0]; };
    Instruction *copy(Instruction *ins, BasicBlock *new_bb);
    // {
    //     LoadInstruction* loadins = dynamic_cast<LoadInstruction*>(ins);
    //     SymbolEntry* dst_se = loadins->getOp()[0]->getSymbolEntry();
    //     SymbolEntry* src_se = loadins->getOp()[1]->getSymbolEntry();
    //     return new LoadInstruction(new Operand(dst_se), new Operand(src_se),
    //     new_bb);
    // }
    void setAsmType(int type){asmtype = type;};
    int getAsmType(){return asmtype;};
};

class StoreInstruction : public Instruction {
    IdentifierSymbolEntry *ident=nullptr;

  public:
    StoreInstruction(Operand *dst_addr, Operand *src,
                     BasicBlock *insert_bb = nullptr, bool assign = 0);
    ~StoreInstruction();
    IdentifierSymbolEntry *getIdentse() { return ident; };
    void setIdentse(IdentifierSymbolEntry *se) { ident = se; };
    void output() const;
    void genMachineCode(AsmBuilder *);
    Operand *getSrc() { return this->operands[1]; };
    Operand *getAddr() { return this->operands[0]; };
    void replaceUse(Operand *old, Operand *new_);
    std::vector<Operand *> getUse() {
        // return std::vector<Operand*>({operands[0], operands[1]});
        return std::vector<Operand *>({operands[1]}); // j0425:给多了
    }
    Instruction *copy1() { return new StoreInstruction(*this); }
    void replaceDef(Operand *new_);
    Operand *getDst() { return nullptr; };
    Instruction *copy(Instruction *ins, BasicBlock *new_bb);
    // {
    //     StoreInstruction* storeins = dynamic_cast<StoreInstruction*>(ins);
    //     SymbolEntry* dst_se = storeins->getOp()[0]->getSymbolEntry();
    //     SymbolEntry* src_se = storeins->getOp()[1]->getSymbolEntry();
    //     return new StoreInstruction(new Operand(dst_se), new Operand(src_se),
    //     new_bb);
    // }

    void setAsmType(int type){asmtype = type;};
    int getAsmType(){return asmtype;};
    private:
        int asmtype = 0;// 初值为0，正常生成汇编。如果赋值为1，生成右端为常数的向量化后的汇编。如果赋值为2，生成右端为数组的向量化后的汇编。
};

class BinaryInstruction : public Instruction {
  public:
    BinaryInstruction(unsigned opcode, Operand *dst, Operand *src1,
                      Operand *src2, BasicBlock *insert_bb = nullptr);
    ~BinaryInstruction();
    void output() const;
    void genMachineCode(AsmBuilder *);
    enum { SUB, ADD, AND, OR, MUL, DIV, MOD };
    std::vector<Operand *> getUse() {
        return std::vector<Operand *>({operands[1], operands[2]});
    }
    Operand *getDef() { return operands[0]; }
    Operand *getDst() { return operands[0]; };
    void replaceUse(Operand *old, Operand *new_);
    void replaceDef(Operand *new_);
    Instruction *copy1() { return new BinaryInstruction(*this); }
    Instruction *copy(Instruction *ins, BasicBlock *new_bb);
    // {
    //     BinaryInstruction* binaryins = dynamic_cast<BinaryInstruction*>(ins);
    //     unsigned mopcode = binaryins->getOpcode();
    //     SymbolEntry* dst_se = binaryins->getOp()[0]->getSymbolEntry();
    //     SymbolEntry* src1_se = binaryins->getOp()[1]->getSymbolEntry();
    //     SymbolEntry* src2_se = binaryins->getOp()[2]->getSymbolEntry();
    //     return new BinaryInstruction(mopcode, new Operand(dst_se), new
    //     Operand(src1_se), new Operand(src2_se), new_bb);
    // }
};

class CmpInstruction : public Instruction {
  public:
    CmpInstruction(unsigned opcode, Operand *dst, Operand *src1, Operand *src2,
                   BasicBlock *insert_bb = nullptr);
    ~CmpInstruction();
    void output() const;
    void genMachineCode(AsmBuilder *);
    enum { E, NE, L, LE, G, GE };
    std::vector<Operand *> getUse() {
        return std::vector<Operand *>({operands[1], operands[2]});
    }
    Operand *getDef() { return operands[0]; }
    Operand *getDst() { return operands[0]; };
    void replaceUse(Operand *old, Operand *new_);
    void replaceDef(Operand *new_);
    Instruction *copy1() { return new CmpInstruction(*this); }
    Instruction *copy(Instruction *ins, BasicBlock *new_bb);
    // {
    //     CmpInstruction* cmpins = dynamic_cast<CmpInstruction*>(ins);
    //     unsigned mopcode = cmpins->getOpcode();
    //     SymbolEntry* dst_se = cmpins->getOp()[0]->getSymbolEntry();
    //     SymbolEntry* src1_se = cmpins->getOp()[1]->getSymbolEntry();
    //     SymbolEntry* src2_se = cmpins->getOp()[2]->getSymbolEntry();
    //     return new CmpInstruction(mopcode, new Operand(dst_se), new
    //     Operand(src1_se), new Operand(src2_se), new_bb);
    // }
};

// unconditional branch
class UncondBrInstruction : public Instruction {
  public:
    UncondBrInstruction(BasicBlock *, BasicBlock *insert_bb = nullptr);
    void output() const;
    void setBranch(BasicBlock *);
    BasicBlock *getBranch();
    void genMachineCode(AsmBuilder *);
    void replaceUse(Operand *old, Operand *new_){};
    Instruction *copy1() { return new UncondBrInstruction(*this); }
    Operand *getDst() { return nullptr; };
    Instruction *copy(Instruction *ins, BasicBlock *new_bb);
    // {
    //     UncondBrInstruction* uncondbrins =
    //     dynamic_cast<UncondBrInstruction*>(ins); BasicBlock* dst_bb =
    //     uncondbrins->getBranch(); return new UncondBrInstruction(dst_bb,
    //     new_bb);
    // }

  protected:
    BasicBlock *branch;
};

// conditional branch
class CondBrInstruction : public Instruction {
  public:
    CondBrInstruction(BasicBlock *, BasicBlock *, Operand *,
                      BasicBlock *insert_bb = nullptr);
    ~CondBrInstruction();
    void output() const;
    void setTrueBranch(BasicBlock *);
    BasicBlock *getTrueBranch();
    void setFalseBranch(BasicBlock *);
    BasicBlock *getFalseBranch();
    void genMachineCode(AsmBuilder *);
    std::vector<Operand *> getUse() {
        return std::vector<Operand *>({operands[0]});
    }
    void replaceUse(Operand *old, Operand *new_);
    Instruction *copy1() { return new CondBrInstruction(*this); }
    Operand *getDst() { return nullptr; };
    Operand *getCond() { return operands[0]; }
    Instruction *copy(Instruction *ins, BasicBlock *new_bb);
    // {
    //     CondBrInstruction* condbrins = dynamic_cast<CondBrInstruction*>(ins);
    //     BasicBlock* dst_truebb = condbrins->getTrueBranch();
    //     BasicBlock* dst_falsebb = condbrins->getFalseBranch();
    //     SymbolEntry* se = condbrins->getOp()[0]->getSymbolEntry();
    //     return new CondBrInstruction(dst_truebb, dst_falsebb, new
    //     Operand(se), new_bb);
    // }

  protected:
    BasicBlock *true_branch;
    BasicBlock *false_branch;
    Operand *cond;
};

class RetInstruction : public Instruction {
  public:
    RetInstruction(Operand *src, BasicBlock *insert_bb = nullptr);
    ~RetInstruction();
    void output() const;
    void genMachineCode(AsmBuilder *);
    void replaceUse(Operand *old, Operand *new_);
    std::vector<Operand *> getUse() {
        if (operands.size())
            return std::vector<Operand *>({operands[0]});
        return std::vector<Operand *>();
    }
    void replaceDef(Operand *new_);
    Instruction *copy1() { return new RetInstruction(*this); }
    Operand *getDst() { return nullptr; };
    Instruction *copy(Instruction *ins, BasicBlock *new_bb);
    // {
    //     RetInstruction* retins = dynamic_cast<RetInstruction*>(ins);
    //     SymbolEntry* se = retins->getOp()[0]->getSymbolEntry();
    //     return new RetInstruction(new Operand(se), new_bb);
    // }
};

class CallInstruction : public Instruction {
  private:
    SymbolEntry *func;

  public:
    Operand *dst;
    CallInstruction(Operand *dst, SymbolEntry *func,
                    std::vector<Operand *> params,
                    BasicBlock *insert_bb = nullptr);
    ~CallInstruction();
    SymbolEntry *getFunc_se() { return func; }
    void output() const;
    SymbolEntry *getFunc() { return func; }
    void genMachineCode(AsmBuilder *);
    void replaceUse(Operand *old, Operand *new_);
    SymbolEntry *getFuncSymbolEntry() { return func; };
    std::vector<Operand *> getParams() {
        std::vector<Operand *> mparams;
        for (int i = 1; i < operands.size(); i++) {
            mparams.push_back(operands[i]);
        }
        return mparams;
    };
    Operand *getDef() { return operands[0]; }
    Operand *getDst() { return operands[0]; };
    void replaceDef(Operand *new_);
    std::vector<Operand *> getUse() {
        std::vector<Operand *> vec;
        for (auto it = operands.begin() + 1; it != operands.end(); it++)
            vec.push_back(*it);
        return vec;
    }
    SymbolEntry *getSymEntry() { return this->func; }
    Instruction *copy1() {
        auto call = new CallInstruction(*this);
        // Todo：添加前驱
        return call;
    }
    Instruction *copy(Instruction *ins, BasicBlock *new_bb);
    // {
    //     CallInstruction* callins = dynamic_cast<CallInstruction*>(ins);
    //     SymbolEntry* dst_se = callins->getOp()[0]->getSymbolEntry();
    //     SymbolEntry* func_se = callins->getFuncSymbolEntry();
    //     std::vector<Operand*> ps = callins->getParams();
    //     return new CallInstruction(new Operand(dst_se), func_se, ps, new_bb);
    // }
};

class FptosiInstruction : public Instruction {
  private:
    Operand *dst;
    Operand *src;

  public:
    FptosiInstruction(Operand *dst, Operand *src,
                      BasicBlock *insert_bb = nullptr);
    ~FptosiInstruction();
    void output() const;
    void genMachineCode(AsmBuilder *);
    // void setDef(Operand* def) {
    //     operands[0] = def;
    //     dst = def;
    //     def->setDef(this);
    // }
    std::vector<Operand *> getUse() {
        return std::vector<Operand *>({operands[1]});
    }
    Operand *getDef() { return operands[0]; }
    Operand *getDst() { return operands[0]; };
    void replaceUse(Operand *old, Operand *new_);
    void replaceDef(Operand *new_);
    Instruction *copy1() { return new FptosiInstruction(*this); }
    Instruction *copy(Instruction *ins, BasicBlock *new_bb);
    // {
    //     FptosiInstruction* fptosiins = dynamic_cast<FptosiInstruction*>(ins);
    //     SymbolEntry* dst_se = fptosiins->getOp()[0]->getSymbolEntry();
    //     SymbolEntry* src_se = fptosiins->getOp()[1]->getSymbolEntry();
    //     return new FptosiInstruction(new Operand(dst_se), new
    //     Operand(src_se), new_bb);
    // }
};

class SitofpInstruction : public Instruction {
  private:
    Operand *dst;
    Operand *src;

  public:
    SitofpInstruction(Operand *dst, Operand *src,
                      BasicBlock *insert_bb = nullptr);
    ~SitofpInstruction();
    void output() const;
    void genMachineCode(AsmBuilder *);
    // void setDef(Operand* def) {
    //     operands[0] = def;
    //     dst = def;
    //     def->setDef(this);
    // }
    std::vector<Operand *> getUse() {
        return std::vector<Operand *>({operands[1]});
    }
    Operand *getDef() { return operands[0]; }
    Operand *getDst() { return operands[0]; };
    void replaceUse(Operand *old, Operand *new_);
    void replaceDef(Operand *new_);
    Instruction *copy1() { return new SitofpInstruction(*this); }
    Instruction *copy(Instruction *ins, BasicBlock *new_bb);
    // {
    //     SitofpInstruction* sitifpins = dynamic_cast<SitofpInstruction*>(ins);
    //     SymbolEntry* dst_se = sitifpins->getOp()[0]->getSymbolEntry();
    //     SymbolEntry* src_se = sitifpins->getOp()[1]->getSymbolEntry();
    //     return new SitofpInstruction(new Operand(dst_se), new
    //     Operand(src_se), new_bb);
    // }
};

class ZextInstruction : public Instruction {
  public:
    ZextInstruction(Operand *dst, Operand *src,
                    BasicBlock *insert_bb = nullptr);
    ~ZextInstruction();
    void output() const;
    void genMachineCode(AsmBuilder *);
    Operand *getDef() { return operands[0]; }
    Operand *getDst() { return operands[0]; };
    void replaceDef(Operand *new_);
    void replaceUse(Operand *old, Operand *new_);
    std::vector<Operand *> getUse() {
        return std::vector<Operand *>({operands[1]});
    }
    Instruction *copy1() { return new ZextInstruction(*this); }
    Instruction *copy(Instruction *ins, BasicBlock *new_bb);
    // {
    //     ZextInstruction* zextins = dynamic_cast<ZextInstruction*>(ins);
    //     SymbolEntry* dst_se = zextins->getOp()[0]->getSymbolEntry();
    //     SymbolEntry* src_se = zextins->getOp()[1]->getSymbolEntry();
    //     return new ZextInstruction(new Operand(dst_se), new Operand(src_se),
    //     new_bb);
    // }
};

class XorInstruction : public Instruction {
  public:
    XorInstruction(Operand *dst, Operand *src, BasicBlock *insert_bb = nullptr);
    ~XorInstruction();
    void output() const;
    void genMachineCode(AsmBuilder *);
    Operand *getDef() { return operands[0]; }
    Operand *getDst() { return operands[0]; };
    void replaceDef(Operand *new_);
    void replaceUse(Operand *old, Operand *new_);
    std::vector<Operand *> getUse() {
        return std::vector<Operand *>({operands[1]});
    }
    Instruction *copy1() { return new XorInstruction(*this); }
    Instruction *copy(Instruction *ins, BasicBlock *new_bb);
    // {
    //     XorInstruction* xorins = dynamic_cast<XorInstruction*>(ins);
    //     SymbolEntry* dst_se = xorins->getOp()[0]->getSymbolEntry();
    //     SymbolEntry* src_se = xorins->getOp()[1]->getSymbolEntry();
    //     return new XorInstruction(new Operand(dst_se), new Operand(src_se),
    //     new_bb);
    // }
};

class GepInstruction : public Instruction {
  private:
    bool paramFirst;
    bool first;
    bool last;
    Operand *init;
    int off;
    // 由于数组初始化使用+4 部分未生成汇编 值编号会重用因此标记
    bool noAsm;
    SymbolEntry * arrSe;

  public:
    GepInstruction(Operand *dst, Operand *arr, Operand *idx,
                   BasicBlock *insert_bb = nullptr, bool paramFirst = false);
    ~GepInstruction();
    void setArrSe(SymbolEntry* se){arrSe=se;};
    SymbolEntry * getArrSe(){return arrSe;};
    void output() const;
    void genMachineCode(AsmBuilder *);
    void setInit1(Operand *init, int off = 0) {
        this->init = init;
        this->off = off;
    };
    void setOff(int f) { off = f; }
    int getOff() { return off; };
    bool getFirst() { return first; };
    void setFirst() { first = true; };
    bool getparamFirst() { return paramFirst; };
    void setparamFirst() { this->paramFirst = true; };
    bool getLast() { return last; };
    void setLast() { last = true; };
    Operand *getInit() const { return init; };
    void setInit(Operand *init) { this->init = init; };
    void setNoAsm(bool na) { noAsm = na; };
    bool hasNoAsm() { return noAsm; }
    Operand *getDef() { return operands[0]; }
    Operand *getDst() { return operands[0]; };
    void replaceDef(Operand *new_);
    void replaceUse(Operand *old, Operand *new_);
    std::vector<Operand *> getUse();
    // std::vector<Operand*> getUse() {

    //     // std::cout<<"check operands[1] 12\n";
    //     // ArrayType* type =
    //     // (ArrayType*)(((PointerType*)(operands[1]->getType()))->getType());
    //     // std::cout<<"origin ayyaytype type = "<<type->toStr()<<"\n";
    //     // Type* elementType = type->getElementType();
    //     // std::cout<<"origin elementtype = "<<elementType<<"\n";

    //     return std::vector<Operand*>({operands[1], operands[2]});
    // }
    // std::vector<Operand*> getUse();
    Instruction *copy(Instruction *ins, BasicBlock *new_bb);
    // {
    //     GepInstruction* gepins = dynamic_cast<GepInstruction*>(ins);
    //     SymbolEntry* dst_se = gepins->getOp()[0]->getSymbolEntry();
    //     SymbolEntry* arr_se = gepins->getOp()[1]->getSymbolEntry();
    //     SymbolEntry* index_se = gepins->getOp()[2]->getSymbolEntry();
    //     return new GepInstruction(new Operand(dst_se), new Operand(arr_se),
    //     new Operand(index_se), new_bb);
    // }
    Instruction *copy1() { return new GepInstruction(*this); }
};
class BitcastInstruction : public Instruction {
  private:
    Operand *dst;
    Operand *src;
    // bool flag;

  public:
    BitcastInstruction(Operand *dst, Operand *src,
                       BasicBlock *insert_bb = nullptr);
    ~BitcastInstruction();
    void output() const;
    void genMachineCode(AsmBuilder *);
    std::vector<Operand *> getUse() {
        return std::vector<Operand *>({operands[1]});
    }
    Operand *getDef() { return operands[0]; }
    Operand *getDst() { return operands[0]; };
    // void setDef(Operand* def) {
    //     operands[0] = def;
    //     dst = def;
    //     def->setDef(this);
    // }
    // void setFlag() { flag = true; }
    // bool getFlag() { return flag; }
    void replaceUse(Operand *old, Operand *new_);
    Instruction *copy1() { return new BitcastInstruction(*this); }
    void replaceDef(Operand *new_);
    Instruction *copy(Instruction *ins, BasicBlock *new_bb);
    // {
    //     BitcastInstruction* bitcastins =
    //     dynamic_cast<BitcastInstruction*>(ins); SymbolEntry* dst_se =
    //     bitcastins->getOp()[0]->getSymbolEntry(); SymbolEntry* src_se =
    //     bitcastins->getOp()[1]->getSymbolEntry(); return new
    //     BitcastInstruction(new Operand(dst_se), new Operand(src_se), new_bb);
    // }
};

class PhiInstruction : public Instruction {
    IdentifierSymbolEntry *ident=nullptr;

  private:
    Operand *originDef;
    Operand *dst;
    std::map<BasicBlock *, Operand *> srcs;
    std::vector<Operand *> phiOperands;
    std::vector<BasicBlock *> phiBBs;

  public:
    IdentifierSymbolEntry *getIdentse() { return ident; };
    void setIdentse(IdentifierSymbolEntry *se) { ident = se; };
    PhiInstruction(Operand *dst, BasicBlock *insert_bb = nullptr);
    ~PhiInstruction();
    std::vector<Operand *> &getPhiOperands_v() { return phiOperands; };
    std::vector<BasicBlock *> &getPhiBBs_v() { return phiBBs; };
    std::vector<Operand *> getPhiOperands() { return phiOperands; };
    std::vector<BasicBlock *> getPhiBBs() { return phiBBs; };
    void output() const;
    Operand *getDef() { return operands[0]; };
    Operand *getDst() { return operands[0]; };
    void genMachineCode(AsmBuilder *builder){};
    void genMachineCodeAfter(AsmBuilder *builder);
    void replaceUse(Operand *old, Operand *new_);
    std::vector<Operand *> getUse() {
        std::vector<Operand *> ret;
        for (auto ope : operands)
            if (ope != operands[0])
                ret.emplace_back(ope);
        return ret;
    }
    void replaceDef(Operand *new_);
    Operand *getOriginDef() { return originDef; }
    void addSrc(BasicBlock *block, Operand *src);
    Operand *getSrc(BasicBlock *block) {
        if (srcs.find(block) != srcs.end())
            return srcs[block];
        return nullptr;
    };
    std::map<BasicBlock *, Operand *> &getSrcs() { return srcs; };
    void changeSrcBlock(BasicBlock *_old, BasicBlock *_new) {
        for (auto src : srcs) {
            if (src.first == _old) {
                Operand *temp = src.second;
                srcs.erase(src.first);
                srcs[_new] = temp;
            }
        }
    }
    Instruction *copy1() { return new PhiInstruction(*this); }
    Instruction *copy(Instruction *ins, BasicBlock *new_bb);
    // {
    //     PhiInstruction* phiins = dynamic_cast<PhiInstruction*>(ins);
    //     SymbolEntry* dst_se = phiins->getOp()[0]->getSymbolEntry();
    //     return new PhiInstruction(new Operand(dst_se), new_bb);
    // }
    void clearSrcs() {
      srcs.clear();
    }
};
#endif